from typing import TYPE_CHECKING
from platon.datastructures import AttributeDict

from platon_aide.base.module import Module
from platon_aide.utils import contract_transaction

if TYPE_CHECKING:
    from platon_aide import Aide


class Delegate(Module):

    def __init__(self, aide: "Aide"):
        super().__init__(aide, module_type='inner-contract')
        self.ADDRESS = self.aide.web3.ppos.delegate.delegateBase.address
        self.REWARD_ADDRESS = self.aide.web3.ppos.delegate.delegateReward.address

    @property
    def staking_block_number(self):
        return self.aide.staking.staking_info.StakingBlockNum

    @contract_transaction()
    def delegate(self,
                 amount=None,
                 balance_type=0,
                 node_id=None,
                 txn=None,
                 private_key=None,
                 ):
        """ 委托节点，以获取节点的奖励分红
        """
        amount = amount or self.aide.economic.delegate_limit
        node_id = node_id or self.aide.node_id
        return self.aide.web3.ppos.delegate.delegate(node_id, balance_type, amount)

    @contract_transaction(1005)
    def withdrew_delegate(self,
                          amount=0,
                          staking_block_identifier=None,
                          node_id=None,
                          txn=None,
                          private_key=None,
                          ):
        """
        撤回对节点的委托，可以撤回部分委托
        注意：因为节点可能进行过多次质押/撤销质押，会使得委托信息遗留，因此撤回委托时必须指定节点质押区块
        """
        node_id = node_id or self.aide.node_id
        amount = amount or self.aide.economic.delegate_limit
        staking_block_identifier = staking_block_identifier or self.aide.staking.staking_info.StakingBlockNum

        return self.aide.web3.ppos.delegate.withdrew_delegate(node_id,
                                                              staking_block_identifier,
                                                              amount,
                                                              )

    @contract_transaction(1006)
    def redeem_delegate(self,
                        txn=None,
                        private_key=None,
                        ):
        """
        领取已经解锁的委托金
        """
        return self.aide.web3.ppos.delegate.redeem_delegate()

    def get_delegate_info(self,
                          address=None,
                          node_id=None,
                          staking_block_identifier=None,
                          ):
        """ 获取地址对某个节点的某次质押的委托信息
        注意：因为节点可能进行过多次质押/撤销质押，会使得委托信息遗留，因此获取委托信息时必须指定节点质押区块
        """
        if self.aide.default_account:
            address = address or self.aide.default_account.address
        node_id = node_id or self.aide.node_id
        staking_block_identifier = staking_block_identifier or self.aide.staking.staking_info.staking_block_number

        delegate_info = self.aide.web3.ppos.delegate.get_delegate_info(address, node_id, staking_block_identifier)
        if delegate_info == 'Query delegate info failed:Delegate info is not found':
            return None
        else:
            return DelegateInfo(delegate_info)

    def get_delegate_lock_info(self,
                               address=None,
                               ):
        """ 获取地址处于锁定期的委托信息
        """
        address = address or self.aide.default_account.address

        delegate_lock_info = self.aide.web3.ppos.delegate.get_delegate_lock_info(address)
        # todo: 根据实际情况补全
        if delegate_lock_info == 'Query delegate info failed:Delegate info is not found':
            return None
        else:
            return delegate_lock_info

    def get_delegate_list(self, address=None):
        """ 获取地址的全部委托信息
        """
        if self.aide.default_account:
            address = address or self.aide.default_account.address
        delegate_list = self.aide.web3.ppos.delegate.get_delegate_list(address)
        if delegate_list == 'Retreiving delegation related mapping failed:RelatedList info is not found':
            return []
        else:
            return [DelegateInfo(delegate_info) for delegate_info in delegate_list]

    @contract_transaction(5000)
    def withdraw_delegate_reward(self,
                                 txn=None,
                                 private_key=None,
                                 ):
        """ 提取委托奖励，会提取委托了的所有节点的委托奖励
        """
        return self.aide.web3.ppos.delegate.withdraw_delegate_reward()

    def get_delegate_reward(self,
                            address=None,
                            node_ids=None
                            ):
        """ 获取委托奖励信息，可以根据节点id过滤
        """
        if self.aide.default_account:
            address = address or self.aide.default_account.address
        node_ids = node_ids or []
        return self.aide.web3.ppos.delegate.get_delegate_reward(address, node_ids)


class DelegateInfo(AttributeDict):
    """ 委托信息的属性字典类
    """
    Addr: str
    NodeId: str
    StakingBlockNum: int
    DelegateEpoch: int
    Released: int
    ReleasedHes: int
    RestrictingPlan: int
    RestrictingPlanHes: int
    CumulativeIncome: int
    LockReleasedHes: int
    LockRestrictingPlanHes: int



import functools
import json
import os
import sys
from os.path import abspath
from typing import cast

from platon import Web3, HTTPProvider, WebsocketProvider, IPCProvider
from platon._utils.threads import Timeout
from platon.datastructures import AttributeDict
from platon.exceptions import ContractLogicError
from platon.types import CodeData

from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.websockets import WebsocketsTransport


def get_web3(uri, chain_id=None, hrp=None, timeout=10, modules=None):
    """ 通过rpc uri，获取web3对象。可以兼容历史platon版本
    """
    if uri.startswith('http'):
        provider = HTTPProvider
    elif uri.startswith('ws'):
        provider = WebsocketProvider
    elif uri.startswith('ipc'):
        provider = IPCProvider
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    with Timeout(timeout) as t:
        while True:
            web3 = Web3(provider(uri), chain_id=chain_id, hrp=hrp, modules=modules)
            if web3.isConnected():
                break
            t.sleep(1)

    return web3


def get_gql(uri):
    """ 通过gql uri，获取gql对象。
    # todo: 增加超时处理
    """
    if uri.startswith('http'):
        transport = AIOHTTPTransport
    elif uri.startswith('ws'):
        transport = WebsocketsTransport
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    return Client(transport=transport(uri), fetch_schema_from_transport=True)


def contract_call(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        return func(self, *args, **kwargs).call()

    return wrapper


# 合约交易装饰器，仅用于接受参数
def contract_transaction(func_id=None, default_txn=None):
    # 实际装饰器
    def decorator(func):

        @functools.wraps(func)
        def wrapper(self, *args, txn: dict = None, private_key=None, **kwargs):
            """
            """
            # 合并交易体
            if txn:
                if default_txn:
                    default_txn.update(txn)
                    txn = default_txn
            else:
                txn = default_txn if default_txn else {}

            # 填充from地址，以免合约交易在预估gas时检验地址失败
            if not txn.get('from'):
                account = self.aide.platon.account.from_key(private_key, hrp=self.aide.hrp) if private_key else self.aide.default_account
                if account:
                    txn['from'] = account.address

            # 构造合约方法对象
            if func.__name__ == 'fit_func':
                # solidity合约方法不传入private key参数，避免abi解析问题
                fn = func(self, *args, **kwargs)
            else:
                # 内置合约有时候需要用到私钥信息，用于生成参数的默认值，如：staking
                fn = func(self, *args, private_key=private_key, **kwargs)

            # 构建合约交易体dict
            try:
                txn = fn.build_transaction(txn)
            except ContractLogicError as e:
                # 预估gas出现合约逻辑错误时，不再报错
                err = str(e)
                # 判断其是否为内置合约错误
                # todo: 优化这一段逻辑
                if err.startswith('inner contract exec failed: '):
                    event = err.split('inner contract exec failed: ')[1]
                    data = json.loads(event.replace('\'', '"'))
                    return cast(CodeData, AttributeDict.recursive(data))
                raise e

            return self._transaction_handler_(txn, func_id=func_id, private_key=private_key)

        return wrapper

    return decorator


def execute_cmd(cmd):
    r = os.popen(cmd)
    out = r.read()
    r.close()
    return out


def mock_duplicate_sign(dtype, sk, blskey, block_number, epoch=0, view_number=0, block_index=0, index=0):
    if sys.platform in "linux,linux2":
        tool_file = abspath("tool/linux/duplicateSign")
        execute_cmd("chmod +x {}".format(tool_file))
    else:
        tool_file = abspath("tool/win/duplicateSign.exe")
    print("{} -dtype={} -sk={} -blskey={} -blockNumber={} -epoch={} -viewNumber={} -blockIndex={} -vindex={}".format(
        tool_file, dtype, sk, blskey, block_number, epoch, view_number, block_index, index))
    output = execute_cmd(
        "{} -dtype={} -sk={} -blskey={} -blockNumber={} -epoch={} -viewNumber={} -blockIndex={} -vindex={}".format(
            tool_file, dtype, sk, blskey, block_number, epoch, view_number, block_index, index))
    print(output)
    if not output:
        raise Exception("unable to use double sign tool")
    return output.strip("\n")



import math
import warnings

from decimal import Decimal
from typing import Literal
from loguru import logger

from platon_aide.base import Module


class Calculator(Module):

    def get_verifier_count(self):
        """ 获取结算周期的验证人数
        备注：目前链上只能获取当前结算周期的验证人数
        """
        verifier_list = self.aide.web3.ppos.staking.get_verifier_list()
        return len(verifier_list)

    def get_block_count(self, node_id, start_bn=None, end_bn=None):
        """ 获取节点出块数
        """
        start_bn = start_bn or 0
        end_bn = end_bn or self.aide.platon.block_number
        if end_bn - start_bn > 1000:
            warnings.warn('too many blocks to analyze, it will be a long wait')

        block_count = 0
        for bn in range(start_bn, end_bn):
            block = self.aide.platon.get_block(bn)
            public_key = self.aide.ec_recover(block)
            if node_id in public_key:
                block_count = block_count + 1

            if bn % 100 == 0:
                logger.info(f'analyzed to {bn}th block, count: {block_count}, waiting...')

        return block_count

    def get_period_info(self,
                        block_number=None,
                        period_type: Literal['round', 'consensus', 'epoch', 'increasing'] = 'epoch'
                        ):
        """ 通过区块高度和周期类型，获取区块所在的周期数和周期结束块高
        """
        if period_type not in ['round', 'consensus', 'epoch', 'increasing']:
            raise ValueError('unknown period type.')

        blocks = {
            'round': self.aide.economic.round_blocks,
            'consensus': self.aide.economic.consensus_blocks,
            'epoch': self.aide.economic.epoch_blocks,
            'increasing': self.aide.economic.increasing_blocks,
        }
        period_blocks = blocks[period_type]

        if not block_number:
            block_number = self.aide.platon.block_number

        period = math.ceil(block_number / period_blocks) or 1
        start_block, end_block = self.get_period_ends(period, period_type)

        return period, start_block, end_block

    def get_period_ends(self,
                        period,
                        period_type: Literal['round', 'consensus', 'epoch', 'increasing'] = 'epoch'
                        ):
        """ 通过周期数和周期类型，获取周期的起始结束块高
        """
        if period_type not in ['round', 'consensus', 'epoch', 'increasing']:
            raise ValueError('unknown period type.')

        blocks = {
            'round': self.aide.economic.round_blocks,
            'consensus': self.aide.economic.consensus_blocks,
            'epoch': self.aide.economic.epoch_blocks,
            'increasing': self.aide.economic.increasing_blocks,
        }
        period_blocks = blocks[period_type]
        start_block, end_block = (period - 1) * period_blocks + 1, period * period_blocks

        return start_block, end_block

    def get_reward_info(self):
        """ 获取当前结算周期的奖励信息
        """
        # todo: 删除该方法
        total_staking_reward = self.aide.web3.ppos.staking.get_staking_reward()
        per_block_reward = self.aide.web3.ppos.staking.get_block_reward()
        return total_staking_reward, per_block_reward

    # def get_node_reward(self, node_id):
    #     """ 即时计算上一个结算周期，节点的总奖励
    #     """
    #     total_staking_reward = self.web3.ppos.staking.get_staking_reward()
    #     verifier_count = self.get_verifier_count()
    #     per_block_reward = self.web3.ppos.staking.get_block_reward()
    #     # 获取上一个结算周期内，节点的出块数
    #     epoch, _, _ = self.get_period_info(self.web3.platon.block_number, 'epoch')
    #     start_bn, end_bn = self.get_period_ends(epoch - 1)
    #     block_count = self.get_block_count(node_id, start_bn, end_bn)
    #     self.calc_node_reward(total_staking_reward,
    #                           verifier_count,
    #                           per_block_reward,
    #                           block_count
    #                           )

    @staticmethod
    def calc_node_reward(total_staking_reward,
                         verifier_count,
                         per_block_reward,
                         block_count,
                         ):
        """ 计算一个结算周期，节点的总奖励（交易gas奖励除外）

        Args:
            total_staking_reward: 结算周期的总质押奖励，可以通过self.get_reward_info获取
            verifier_count: 结算周期的验证人数，可以通过self.get_verifier_count获取
            per_block_reward: 结算周期的出块奖励，可以通过self.get_reward_info获取
            block_count: 节点在结算周期的出块数，可以通过self.get_block_count获取
        """
        staking_reward = int(Decimal(total_staking_reward) / Decimal(verifier_count))
        block_reward = int(Decimal(per_block_reward) * Decimal(block_count))
        node_reward = staking_reward + block_reward
        return node_reward

    # def get_staking_reward(self, node_id):
    #     """ 即时计算上一个结算周期，节点的质押奖励
    #     """
    #     total_staking_reward = self.aide.web3.ppos.staking.get_staking_reward()
    #     verifier_count = self.get_verifier_count()
    #     per_block_reward = self.aide.web3.ppos.staking.get_block_reward()
    #     # 获取上一个结算周期内，节点的出块数
    #     epoch, _, _ = self.get_period_info(self.web3.platon.block_number, 'epoch')
    #     start_bn, end_bn = self.get_period_ends(epoch - 1)
    #     block_count = self.get_block_count(node_id, start_bn, end_bn)
    #
    #     node_reward_ratio = self._staking.get_candidate_info(node_id).RewardPer
    #
    #     return self.calc_staking_reward(total_staking_reward,
    #                                     verifier_count,
    #                                     per_block_reward,
    #                                     block_count,
    #                                     node_reward_ratio,
    #                                     )

    @staticmethod
    def calc_staking_reward(total_staking_reward,
                            verifier_count,
                            per_block_reward,
                            block_count,
                            node_reward_ratio,
                            ):
        """
        计算一个结算周期，节点的质押奖励

        Args:
            total_staking_reward: 结算周期的总质押奖励，可以通过self.get_reward_info获取
            verifier_count: 结算周期的验证人数，可以通过self.get_verifier_count获取
            per_block_reward: 结算周期的出块奖励，可以通过self.get_reward_info获取
            block_count: 节点在结算周期的出块数，可以通过self.get_block_count获取
            node_reward_ratio: 节点在结算周期的委托分红比例
        """
        node_reward = Calculator.calc_node_reward(total_staking_reward,
                                                  verifier_count,
                                                  per_block_reward,
                                                  block_count,
                                                  )

        delegate_reward = Calculator.calc_delegate_reward(total_staking_reward,
                                                          verifier_count,
                                                          per_block_reward,
                                                          block_count,
                                                          node_reward_ratio,
                                                          )

        staking_reward = node_reward - delegate_reward
        return staking_reward

    # def get_delegate_reward(self,
    #                         node_id,
    #                         delegate_amount,
    #                         delegate_total_amount,
    #                         ):
    #     """
    #     即时计算上一个结算周期，账户在节点的委托奖励
    #
    #     Args:
    #         node_id: 委托节点ID
    #         delegate_amount: 结算周期内，账户在节点的锁定期委托总额
    #         delegate_total_amount: 节点在该结算周期的锁定期委托总额
    #     """
    #     total_staking_reward = self.web3.ppos.staking.get_staking_reward()
    #     verifier_count = self.get_verifier_count()
    #     per_block_reward = self.web3.ppos.staking.get_block_reward()
    #     # 获取上一个结算周期内，委托节点的出块数
    #     epoch, _, _ = self.get_period_info(self.web3.platon.block_number, 'epoch')
    #     start_bn, end_bn = self.get_period_ends(epoch - 1)
    #     block_count = self.get_block_count(node_id, start_bn, end_bn)
    #
    #     node_reward_ratio = self._staking.get_candidate_info(node_id).RewardPer
    #
    #     return self.calc_delegate_reward(total_staking_reward,
    #                                      verifier_count,
    #                                      per_block_reward,
    #                                      block_count,
    #                                      node_reward_ratio,
    #                                      delegate_total_amount,
    #                                      delegate_amount,
    #                                      )

    @staticmethod
    def calc_delegate_reward(total_staking_reward,
                             verifier_count,
                             per_block_reward,
                             block_count,
                             node_reward_ratio,
                             delegate_total_amount=None,
                             delegate_amount=None,
                             ):
        """
        计算一个结算周期内，账户在节点的委托奖励

        Args:
            total_staking_reward: 结算周期的总质押奖励，可以通过self.get_reward_info获取
            verifier_count: 结算周期的验证人数，可以通过self.get_verifier_count获取
            per_block_reward: 结算周期的出块奖励，可以通过self.get_reward_info获取
            block_count: 节点在结算周期的出块数，可以通过self.get_block_count获取
            node_reward_ratio: 节点在结算周期的委托分红比例
            delegate_total_amount: 节点在结算周期的锁定期委托总额
            delegate_amount: 结算周期内，账户对节点的锁定期委托总额
        """
        dividend_ratio = Decimal(node_reward_ratio) / Decimal(10000)
        # 按照底层算法，计算质押奖励分红
        staking_reward = int(Decimal(total_staking_reward) / Decimal(verifier_count))
        staking_dividends = int(Decimal(staking_reward) * dividend_ratio)
        # 按照底层算法，计算出块奖励分红
        per_block_delegate_reward = Decimal(per_block_reward) * dividend_ratio
        block_dividends = per_block_delegate_reward * Decimal(block_count)
        # 计算节点所有委托分红
        all_delegate_reward = int(staking_dividends + block_dividends)
        if not delegate_total_amount and not delegate_amount:
            return all_delegate_reward
        # 计算账户的委托分红
        delegate_reward = math.floor(Decimal(all_delegate_reward) * Decimal(delegate_amount) / Decimal(delegate_total_amount))
        return delegate_reward

    def calc_report_multi_sign_reward(self, staking_amount):
        """ 计算举报双签的奖励
        """
        slashing_ratio = self.aide.economic.slashing.slashFractionDuplicateSign
        slashing_amount = Decimal(staking_amount) * (Decimal(slashing_ratio) / 10000)

        reward_ratio = self.aide.economic.slashing.duplicateSignReportReward
        report_reward = slashing_amount * (Decimal(reward_ratio) / 100)

        to_incentive_pool_amount = slashing_amount - report_reward
        return int(report_reward), int(to_incentive_pool_amount)
